//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.2
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Version 0.1: 1999/05/05
// Copyright (C) 1999 Jun-ichi Yoshida and Ikki Ohmukai, All rights reserved.
// Version 1.0: 2001/12/13
// Copyright (C) 2001 Jun-ichi Yoshida, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
// Version 1.1: 2002/01/17
// Copyright (C) 2002 Jun-ichi Yoshida, Masaki Sano, Tomoyuki Hiroyasu and Mitsunori Miki, All rights reserved.
// Version 1.2: 2002/06/04
// Copyright (C) 2002 Jun-ichi Yoshida, Masaki Sano, Tomoyuki Hiroyasu and Mitsunori Miki, All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notce and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright
//    notice and this list of conditions in the documentation and/or other
//    materials provided with the distribution.
// 3. The names of copyrighters of this software  should not be used to
//    endorse or promote products derived from this software without specific
//    prior written permission.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k_mpi.cpp
//-----------------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cfloat>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
using namespace std;

#include <mpi.h>
#include "ga2k_mpi.h"
#include "randring.h"



//=============================================================================
//functions
//=============================================================================

static int send_gene(const Gene& gene, int dest)
{
	void *buf = (void*)(gene.bitString.begin());
	int count = gene.bitString.size();
	int tag = 1;

	return MPI_Send(buf, count, MPI_CHAR, dest, tag, MPI_COMM_WORLD);
}

static int recv_gene(Gene& gene)
{
	void *buf = (void*)(gene.bitString.begin());
	int count = gene.bitString.size();
	MPI_Status stat;

	return MPI_Recv(buf, count, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG,
		MPI_COMM_WORLD, &stat);
}


//=============================================================================
//class MPIGlobe
//=============================================================================

MPIGlobe::MPIGlobe(int *pargc, char*** pargv) : destination(2)
{
	MPI_Init(pargc, pargv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
	int namelen = 0;
	MPI_Get_processor_name(hostname, &namelen);

	fprintf(stderr, "Process %d on %s\n", myRank, hostname);
	root = 0;
	synchronize();
}

MPIGlobe::~MPIGlobe()
{
	MPI_Finalize();
}

void MPIGlobe::initialize()
{
// Default settings

	fTotalPopSize = false;
	totalPopulationSize = 400;

	island_0.islandId		= myRank;

	// parameters for chromosome
	island_0.geneLength		= 100;
	island_0.variableLength	= 10;
	island_0.numVariable		= 10;

	fInitialPopulation			= 0; // 0: random,  1: from file(design variables)

	// parameters for objective function
	island_0.functionType		= 0;
	island_0.openMax			= 5.12;
	island_0.closedMin			= -5.12;
	island_0.delta				= 0.01;
	island_0.finValue			= -0.00000001;

	// canonical GA parameters
	island_0.populationSize		= 10; // in each subpopulations
	island_0.crossoverRate		= 1.0;
	island_0.mutationRate		= 0.01;
	island_0.numElite			= 5;
	island_0.tournamentSize    = 4;
	island_0.selectionType		= 0; // 0: tournament,  1: roulette

	// DGA parameters
	island_0.migrationRate		= 0.5;
	numIsland					= nProcs;
	migrationInterval			= 5;


	island_0.numCrossoverPoint	= 2;
	island_0.typeOfEncode		= 0;


	island_0.numEvaluate		= 0;
	maxGeneration		= 1000;


	maxTrial			= 1;
	sumMaxFitness		= 0;
	sumGeneration		= 0;
	numConvergent		= 0;
	sumTotalEvaluate	= 0;
	fStop				= true;
	fOutput				= false;
}

void MPIGlobe::initialize(int argc, char *argv[])
{
	initialize();

	// command line
	for(int i = 1; i < argc; i++)
	{
		const char *p = argv[i] + 1;
		switch(*argv[i])
		{

		case 'b':
			island_0.geneLength = atoi(p);
			island_0.numVariable = island_0.geneLength / island_0.variableLength;
			break;

		case 'i':
			numIsland = atoi(p);
			break;

		case 'c':
			island_0.crossoverRate = atof(p);
			break;

		case 'f':
			island_0.functionType = atoi(p);
			if(atoi(p) == 0){
				island_0.closedMin = -5.12;
				island_0.openMax = 5.12;
				island_0.delta = 0.01;
			}
			else if(atoi(p) == 1 || atoi(p) == 2){
				island_0.closedMin = -512;
				island_0.openMax = 512;
				island_0.delta = 1;
			}
			else if(atoi(p) == 3){
				island_0.closedMin = -2.048;
				island_0.openMax = 2.048;
				island_0.delta = 0.004;
			}
			else if(atoi(p) == 4){
				island_0.closedMin = -64;
				island_0.openMax = 64;
				island_0.delta = 0.125;
			}
			break;

		case 'I':
			migrationInterval = atoi(p);
			break;

		case 'm':
			island_0.mutationRate = atof(p);
			break;

		case 'M':
			island_0.migrationRate = atof(p);
			break;

		case 'g':
			maxGeneration = atoi(p);
			break;

		case 'o':
			fOutput = true;
			break;

		case 'p':
			island_0.populationSize = atoi(p);
			island_0.numElite = island_0.populationSize / 2;
			break;

		case 'P':
			totalPopulationSize = atoi(p);
			fTotalPopSize = true;
			break;

		case 'z':
			maxTrial = atoi(p);
			break;

		case 'S':
			// 0: random,  1: from file(design variables)
			fInitialPopulation	= atoi(p);
			break;

		case 's':
			// 0: tournament,  1: roulette
			island_0.selectionType = atoi(p);
			break;
		}
	}
}

void MPIGlobe::establish()
{
	if(fTotalPopSize)
	{
		island_0.populationSize = totalPopulationSize / numIsland;
		island_0.numElite = island_0.populationSize / 2;
	}
	island_0.migrantSize	= static_cast<int>(island_0.populationSize * island_0.migrationRate);
	island_0.nativeSize	= island_0.populationSize - island_0.migrantSize;

	populationSize	= island_0.populationSize;
	nativeSize		= island_0.nativeSize;
	migrantSize		= island_0.migrantSize;

	maxFitness		= -DBL_MAX;
	convGeneration	= maxGeneration;
	fConv			= false;
	totalEvaluate	= 0;

	// broadcast seed for random ring
	int seed = genrand_int(INT_MAX);
	MPI_Bcast(&seed, 1, MPI_INT, root, MPI_COMM_WORLD);
	destination.srand(seed);
	destination.resize(numIsland);


	ifstream fin;
	if(fInitialPopulation == 1){
		// ݌vϐl̃t@CI[v
		fin.open(INITIAL_POPULATION_FILE);
		if(!fin.is_open()){
			cerr << __FILE__ << ":(" << __LINE__ << ") can't open file "
				<< INITIAL_POPULATION_FILE << endl;
			exit(1);
		}
	}

	island = island_0;
	island.initialize(myRank, nProcs, fin);

	fin.close();

	synchronize();
}

void MPIGlobe::emigrate()
{
	if(migrantSize == 0 || nProcs < 2)
		return;


	// siڏZg|W̓_Oj
	destination.shuffle();

	// ڏZ̂̓_
	random_shuffle(island.gene.begin(), island.gene.end(), genrand_int);

	// s̓ migrant Ɍ̂Rs[
	int dest = destination[myRank];
	for(int j = 0; j < migrantSize; j++){
		send_gene(island.gene[nativeSize + j], dest);
		recv_gene(island.migrant[j]);
	}
}

void MPIGlobe::execute()
{
	for(generation = 0; generation < maxGeneration; generation++)
	{
		if(generation % 10 == 0)
			if(myRank == root)
				cerr << "." << flush;

		if(generation % migrationInterval == 0){
			emigrate();
			island.immigrate();
		}

		island.crossover();
		island.mutate();
		island.evaluate();
		island.elitePreservation();

		setMaxFitness();

		island.reproduce();

		if(!fConv && maxFitness > island_0.finValue)
			finalize(true);

		sumFitness[generation] += maxFitness;
	}
	finalize(false);
}


void MPIGlobe::finalize(bool b)
{
	if(b)
	{
		maxFitness = 0.000000;
		sumMaxFitness += maxFitness;
		convGeneration = generation;
		sumGeneration += convGeneration;
		numConvergent++;
		if(myRank == root) cerr << "*" << flush;
		fConv = true;

		if(fStop)
			for(; generation < maxGeneration; generation++)
				sumFitness[generation] += maxFitness;

		MPI_Reduce(&(island.numEvaluate), &(totalEvaluate), 1, MPI_INT, MPI_SUM, root, MPI_COMM_WORLD);
		MPI_Bcast(&totalEvaluate, 1, MPI_INT, root, MPI_COMM_WORLD);

		sumTotalEvaluate += totalEvaluate;
	}
	else if(!b && !fConv)
	{
		sumMaxFitness += maxFitness;
		MPI_Reduce(&(island.numEvaluate), &(totalEvaluate), 1, MPI_INT, MPI_SUM, root, MPI_COMM_WORLD);
		MPI_Bcast(&totalEvaluate, 1, MPI_INT, root, MPI_COMM_WORLD);
		sumTotalEvaluate += totalEvaluate;
	}
}

//sOɃp[^cerrŏo͂
void MPIGlobe::showParameters()
{
	if(myRank == root){
		cerr << "\nPopulation     : " << island_0.populationSize * numIsland << endl;
		cerr << "Num island     : " << numIsland << endl;
		cerr << "Pop/island     : " << island_0.populationSize << endl;
		cerr << "init population: " << ((fInitialPopulation == 0) ? "random" : "from file") << endl;
		cerr << "Elite/island   : " << island_0.numElite << endl;
		cerr << "MigRate        : " << island_0.migrationRate << endl;
		cerr << "MigInterval    : " << migrationInterval << endl ;
		cerr << "Pc             : " << island_0.crossoverRate << endl;
		cerr << "Pm             : " << island_0.mutationRate << endl;;
		cerr << "MaxGeneration  : " << maxGeneration << endl;

		//֐Ɋւ
		cerr << "Function       : " <<  "f" << island_0.functionType << endl ;
		cerr << "\n";
		cerr << flush;
	}
	synchronize();
}

void MPIGlobe::synchronize() const
{
	MPI_Barrier(MPI_COMM_WORLD);
}

void MPIGlobe::setMaxFitness()
{
	MPI_Reduce(&(island.bestFitness), &maxFitness, 1, MPI_DOUBLE,
		MPI_MAX, root, MPI_COMM_WORLD);
	MPI_Bcast(&maxFitness, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
}

